﻿using System.CommandLine;
using System.CommandLine.Builder;
using System.CommandLine.Help;
using System.CommandLine.Invocation;
using System.CommandLine.Parsing;

namespace Ozeki
{
    internal class HFDownloadRootCommand : RootCommand
    {
        private const string HelpMessage =
@"OZEKI HFDownloader v1.0.0

Description:
    This tool makes it possible to download files from HuggingFace.co

Usage:
    HFDownload <url> [options]

Parameters:
    <url>   the URL of the HuggingFace repository or file to be downloaded
    -f      force download (overwrite the file if exists) [default: False]
    -t <t>  HuggingFace access token to login
    -n <n>  number of threads to use [default: 5]
    -o <o>  output directory [default: C:\AIModels]
    -s <s>  maximum download speed in megabytes/sec [default: unlimited]
    -r <r>  number of retries [default: 20]
    -?      help
    -v      version information

Examples:
    Download full repository:
        hfdownload.exe https://huggingface.co/nvidia/Llama-3.1-Nemotron-70B-Instruct-HF/
    Download single file from repository to custom location:
        hfdownload.exe /nvidia/Llama-3.1-Nemotron-70B-Instruct-HF/tokenizer.json -o C:/AIModels

For more information, visit: https://ozeki-ai-server.com/p_8742-hfdownloader.html
";


        private readonly Argument<string?> argUrl = new Argument<string?>("url", () => null);
        private readonly Option<bool> optForce = new Option<bool>("-f", () => false, "specifies if the download should be forced");
        private readonly Option<string> optAccessToken = new Option<string>("-t", "specifies the HuggingFace Access Token");
        private readonly Option<int> optThreadNumber = new Option<int>("-n", () => 5, "specifies how many threads can be used by the downloading process");
        private readonly Option<string> optOutputLocation = new Option<string>("-o", getDefaultOutputLocation, "specifies where the files should be downloaded");
        private readonly Option<string> optDownloadSpeedInBits = new Option<string>("-s");
        private readonly Option<int> optNumberOfRetries = new Option<int>("-r", () => 20);

        private static string getDefaultOutputLocation()
        {
            if (Environment.OSVersion.Platform == PlatformID.Win32NT)
            {
                return "C:\\AIModels";
            }
            else
            {
                return "/AIModels";
            }
        }

        private HFDownloadRootCommand() {
            Description = "This tool makes it possible to download files from HuggingFace.co";
            Add(argUrl);
            Add(optForce);
            Add(optAccessToken);
            Add(optThreadNumber);
            Add(optOutputLocation);
            Add(optDownloadSpeedInBits);
            Add(optNumberOfRetries);
            this.SetHandler(parseArguments);
            this.AddValidator(validate);
        }

        public static Parser Create()
        {
            var rootCommand = new HFDownloadRootCommand();
            var parser = new CommandLineBuilder(rootCommand)
               .UseHelp("-?")
               .UseVersionOption("-v")
               .UseHelp(ctx =>
               {
                   ctx.HelpBuilder.CustomizeLayout(_ => HelpBuilder.Default.GetLayout().Where((x) => false));
                   ctx.Output.WriteLine(HelpMessage);
               })
               .UseEnvironmentVariableDirective()
               .UseParseDirective()
               .UseSuggestDirective()
               //.RegisterWithDotnetSuggest()
               .UseTypoCorrections()
               .UseParseErrorReporting()
               .UseExceptionHandler()
               .CancelOnProcessTermination()
               .Build();

            return parser;
        }
        private void validate(CommandResult symbolResult)
        {
            var url = symbolResult.GetValueForArgument(argUrl);
            var output = symbolResult.GetValueForOption(optOutputLocation);
            var threadnumber = symbolResult.GetValueForOption(optThreadNumber);
            var downloadSpeed = symbolResult.GetValueForOption(optDownloadSpeedInBits);
            var numberOfRetries = symbolResult.GetValueForOption(optNumberOfRetries);

            if (Uri.TryCreate("https://huggingface.co/" + url, UriKind.Absolute, out var fullUrl) ||  Uri.TryCreate(url, UriKind.Absolute, out fullUrl))
            {
                if (fullUrl.Scheme != "https")
                {
                    symbolResult.ErrorMessage = "Invalid URL: Scheme is not https";
                    return;
                }
                
                if (fullUrl.Host != "huggingface.co")
                {
                    symbolResult.ErrorMessage = "Invalid URL: Host is not huggingface.co";
                    return;
                }

                if(fullUrl.AbsolutePath == "")
                {
                    symbolResult.ErrorMessage = "Invalid URL: No path specified";
                    return;
                }
            }
            else
            {
                symbolResult.ErrorMessage = "Invalid URL: Specified URL can't be parsed";
                return;
            }

            if (downloadSpeed != null && !parseDownloadSpeed(downloadSpeed, out var speedInBytes))
            {
                symbolResult.ErrorMessage = "Invalid Download Speed: Specified Download Speed can't be parsed";
                return;
            }


            if(threadnumber < 1)
            {
                symbolResult.ErrorMessage = "Invalid Thread Number: Number of threads must be a positive non-zero number.";
                return;
            }
            if(numberOfRetries < 1)
            {
                symbolResult.ErrorMessage = "Invalid Retry Number: Number of retries must be a positive non-zero number.";
                return;
            }
            if(output == null)
            {
                return;
            }
            try
            {
                Directory.CreateDirectory(output);
            }
            catch (ArgumentException)
            {
                symbolResult.ErrorMessage = "Invalid Output Path: Specified path not valid: " + output;
                return;
            }
            catch (UnauthorizedAccessException)
            {
                symbolResult.ErrorMessage = "Invalid Output Path: No permission to access " + output;
                return;
            }
            catch (Exception ex)
            {
                symbolResult.ErrorMessage = "Invalid Output Path: " + ex.Message;
                return;
            }
        }    

        private void parseArguments(InvocationContext context)
        {
            var url = context.ParseResult.GetValueForArgument(argUrl);
            var output = context.ParseResult.GetValueForOption(optOutputLocation);
            var force = context.ParseResult.GetValueForOption(optForce);
            var accesstoken = context.ParseResult.GetValueForOption(optAccessToken);
            var threadnumber = context.ParseResult.GetValueForOption(optThreadNumber);
            var downloadSpeed = context.ParseResult.GetValueForOption(optDownloadSpeedInBits);
            var numberOfTries = context.ParseResult.GetValueForOption(optNumberOfRetries);
            if (url == null)
            {
                Console.WriteLine(HelpMessage);
                return;
            }
            if(output == null)
            {
                Console.WriteLine("Invalid Output Path: No output path specified");
                return;
            }
            long speedInBytes = -1;
            if (downloadSpeed != null)
            {
               parseDownloadSpeed(downloadSpeed, out speedInBytes);
            }
            
            var param = new OzHfDownloadParams(url, accesstoken, threadnumber, force, output, numberOfTries, speedInBytes == -1 ? null : speedInBytes);
            var downloadUI = new DownloadUI();
            downloadUI.Initialize();
            var success = OzHfDownloader.Download(param, out var error, downloadUI);
            downloadUI.Stop();
            if (!success) {
                Console.WriteLine(error);
                return;
            }
            Console.WriteLine("HFDownloader ran successfully.");
        }

        public static bool parseDownloadSpeed(string downloadSpeed, out long speedInBytes)
        {
            
            if (long.TryParse(downloadSpeed,out var speed))
            {
                speedInBytes = speed * 1000 * 1000;
                return true;
            }
            string[] SizeSuffixes = ["b", "B", "Kb", "kb", "Mb", "mb", "Gb", "gb", "Tb", "tb", "Pb", "pb", "Eb", "eb", "Zb", "zb", "Yb", "yb"];
            for (int i = 1; i < SizeSuffixes.Length; i++)
            {
                if (downloadSpeed.EndsWith(SizeSuffixes[i]))
                {
                    if (long.TryParse(downloadSpeed[..^2], out speed))
                    {
                        speedInBytes = speed * (long)Math.Pow(1000, i);
                        return true;
                    }
                }
            }            

            speedInBytes = 0;
            return false;
        }
    }
}
