﻿namespace Ozeki
{
    class OzHfFileSelector
    {
        private List<Option> _options = new List<Option>();
        private int _currentPage = 0;
        private int _currentSelection = 0;
        private int _windowHeight = Console.WindowHeight - 3;
        private long _avalaibleSpace = 0;
        private long _selectedSpace = 0;


        private void renderPage()
        {
            renderHead();
            renderStats();
            var page = _options.Skip(calculateAbsoluteIndex(0)).Take(_windowHeight).ToList();
            if (_currentPage == maxPage)
            {
                page = _options.Skip(calculateAbsoluteIndex(0)).ToList();
            }
            for (int i = 0; i < page.Count; i++)
            {
                renderLine(i);
            }
        }

        private int calculateAbsoluteIndex(int index) => _currentPage * _windowHeight + index;
        private int maxPage => (int)Math.Floor((double)_options.Count / _windowHeight);
        private int lastPageCount => _options.Count - _windowHeight * maxPage;
        public IEnumerable<string> Select()
        {
            bool running = true;
            renderPage();
            while (running)
            {
                var key = Console.KeyAvailable ? Console.ReadKey(true).Key : ConsoleKey.Home;
                switch (key)
                {
                    case ConsoleKey.Enter:
                        return _options.Where(x => x.Selected).Select(x => x.Text);

                    case ConsoleKey.Spacebar:
                        select();
                        break;

                    case ConsoleKey.DownArrow:
                        down();
                        break;

                    case ConsoleKey.UpArrow:
                        up();
                        break;

                    default:
                        break;
                }
            }
            return [];
        }

        private void select()
        {
            var index = calculateAbsoluteIndex(_currentSelection);
            var filepath = _options[index].Manifest.OutputPath;
            long alreadyDownloaded = 0;
            if (File.Exists(filepath))
            {
                alreadyDownloaded = new FileInfo(filepath).Length;
            }

            if (_avalaibleSpace - (_options[index].Manifest.Size + _selectedSpace) < 0)
            {
                renderLine(_currentSelection, "Not enough space remaining to select.");
                return;
            }

            if (_options[index].Selected)
            {
                _selectedSpace -= _options[index].Manifest.Size - alreadyDownloaded;
            }
            else
            {
                _selectedSpace += _options[index].Manifest.Size - alreadyDownloaded;
            }
            _options[index].Selected = !_options[index].Selected;
            renderStats();
            renderLine(_currentSelection);
        }

        private void down()
        {
            if (_currentPage == maxPage && _currentSelection != lastPageCount - 1)
            {

                _currentSelection++;
                renderLine(_currentSelection - 1);
                renderLine(_currentSelection);
            }
            else if (_currentSelection + 1 != _windowHeight && _currentPage != maxPage)
            {
                _currentSelection++;
                renderLine(_currentSelection - 1);
                renderLine(_currentSelection);
            }
            else if (_currentSelection + 1 == _windowHeight && _currentPage != maxPage)
            {
                _currentPage++;
                Console.Clear();
                _currentSelection = 0;
                renderPage();
            }
        }

        private void up()
        {
            if (_currentSelection != 0)
            {
                _currentSelection--;
                renderLine(_currentSelection + 1);
                renderLine(_currentSelection);
            }
            else if (_currentSelection == 0 && _currentPage != 0)
            {
                _currentSelection = _windowHeight - 1;
                _currentPage--;
                Console.Clear();
                renderPage();
            }
        }

        private void renderHead()
        {
            Console.SetCursorPosition(0, 0);
            Console.WriteLine($"Page: {(_options.Count / _windowHeight) + 1}/{_currentPage + 1} Arrows: navigate between items; Space: select item; Enter: finish selection");
        }

        private void renderStats()
        {
            var selected = _options.Where(x => x.Selected);
            Console.SetCursorPosition(0, 1);
            var totalSelected = selected.Count();
            var totalSize = selected.Sum(x => x.Manifest.Size);
            var line = $"Files selected: {totalSelected} Total download size: {OzHfDownloadDetails.FileSizeToString(totalSize)}";
            Console.Write(line);
            for (int i = line.Length - 1; i < Console.WindowWidth - 1; i++)
            {
                Console.Write(" ");
            }
            Console.WriteLine();
            var secondline = $"Space remaining after download: { OzHfDownloadDetails.FileSizeToString(_avalaibleSpace - _selectedSpace)}";
            Console.Write(secondline);
            for (int i = secondline.Length - 1; i < Console.WindowWidth - 1; i++)
            {
                Console.Write(" ");
            }
        }

        private void renderLine(int index, string? message = null)
        {
            if (index == _currentSelection) Console.ForegroundColor = ConsoleColor.Blue;
            Console.SetCursorPosition(0, (int)index + 3);
            Console.Write(_options[calculateAbsoluteIndex(index)]);
            Console.ForegroundColor = ConsoleColor.White;
            if (message != null) Console.Write(message);
        }
        public OzHfFileSelector(IEnumerable<OzHfDownloadManifest> manifests, IEnumerable<string> selected, long space)
        {
            var outputlocation = "";
            foreach (var manifest in manifests)
            {
                var option = new Option(manifest.FileName) { Manifest = manifest };
                _options.Add(option);
                outputlocation = manifest.OutputPath;
            }
            _avalaibleSpace = new DriveInfo(Directory.GetDirectoryRoot(outputlocation)).AvailableFreeSpace; 

            foreach (var select in selected)
            {
                _options.Find(x => x.Manifest.FileName == select).Selected = true;               
            }

            _selectedSpace = _options.Where(x => x.Selected).Sum(x => x.Manifest.Size);
            _selectedSpace -= _options.Where(x => x.Selected).Sum(x => File.Exists(x.Manifest.OutputPath) ? new FileInfo(x.Manifest.OutputPath).Length : 0);

        }

        protected class Option
        {
            public bool Selected { get; set; }
            public string Text { get; set; }
            public OzHfDownloadManifest Manifest { get; set; }

            public Option(string text)
            {
                Selected = false;
                Text = text;
            }

            public override string ToString()
            {
                return $"[{(Selected ? '*' : ' ')}] {Text} {OzHfDownloadDetails.FileSizeToString(Manifest.Size)}";
            }
        }
    }
}
