﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reactive.Concurrency;
using System.Text;
using System.Threading.Tasks;

namespace Ozeki
{

    public class OzHfThrottledStream : Stream
    {
        private readonly Stream _parent;
        private long? _maxBitsPerSecond;
        private readonly IScheduler _scheduler;
        private readonly IStopwatch _stopwatch;

        private long _bitsProcessed;


        public OzHfThrottledStream(Stream parent, long? maxBitsPerSecond, IScheduler scheduler)
        {
            this._maxBitsPerSecond = maxBitsPerSecond;
            this._parent = parent;
            this._scheduler = scheduler;
            _stopwatch = scheduler.StartStopwatch();
            _bitsProcessed = 0;
        }

        public OzHfThrottledStream(Stream parent, long? maxBytesPerSecond)
            : this(parent, maxBytesPerSecond, Scheduler.Immediate)
        {
        }

        protected void Throttle(int bytes)
        {
            if (this._maxBitsPerSecond == null)
            {
                return;
            }
            var maxBitsPerSecond = this._maxBitsPerSecond.Value;

            _bitsProcessed += bytes * 8;
            var targetTime = TimeSpan.FromMilliseconds((double)_bitsProcessed / (maxBitsPerSecond / 1000));
            var actualTime = _stopwatch.Elapsed;
            var sleep = TimeSpan.FromMilliseconds(targetTime.TotalMilliseconds - actualTime.TotalMilliseconds);
            if (sleep > TimeSpan.Zero)
            {
                using (var waitHandle = new AutoResetEvent(initialState: false))
                {
                    _scheduler.Sleep(sleep).GetAwaiter().OnCompleted(() => waitHandle.Set());
                    waitHandle.WaitOne();
                }
            }
        }

        public override bool CanRead
        {
            get { return _parent.CanRead; }
        }

        public override bool CanSeek
        {
            get { return _parent.CanSeek; }
        }

        public override bool CanWrite
        {
            get { return _parent.CanWrite; }
        }

        public override void Flush()
        {
            _parent.Flush();
        }

        public override long Length
        {
            get { return _parent.Length; }
        }

        public override long Position
        {
            get
            {
                return _parent.Position;
            }
            set
            {
                _parent.Position = value;
            }
        }

        public long? MaxBitsPerSecond { get => _maxBitsPerSecond; set => _maxBitsPerSecond = value; }

        public override int Read(byte[] buffer, int offset, int count)
        {
            var read = _parent.Read(buffer, offset, count);          
            Throttle(read);            
            return read;
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            return _parent.Seek(offset, origin);
        }

        public override void SetLength(long value)
        {
            _parent.SetLength(value);
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            Throttle(count);
            _parent.Write(buffer, offset, count);
        }
    }
}

