﻿using System.Net.Http.Headers;
using System.Net;
using System.Text.Json;
using System.Net.NetworkInformation;

namespace Ozeki
{
    internal static class OzHfNetworkOperation
    {
        public static bool IsHuggingFaceReachable()
        {
            try
            {
                using (Ping pinger = new Ping())
                {
                    PingReply reply = pinger.Send("huggingface.co");
                    return reply.Status == IPStatus.Success;
                }
            }
            catch (PingException)
            {
                return false;
            }
        }


        public static long GetContentLength(string url, string? accesskey = null)
        {
            int i = 0;
            while (i < 5)
            {
                try
                {
                    HttpClient client = OzHfHTTPClient.Get();
                    var request = new HttpRequestMessage(HttpMethod.Head, url);
                    if(accesskey != null)
                    {
                        request.Headers.Authorization = new AuthenticationHeaderValue("Bearer", accesskey);
                    }    
                    var response = client.Send(request);
                    var length = response.Content.Headers.ContentLength;
                    if (length == 43) throw new Exception();
                    return length ?? 0;
                }
                catch (Exception e)
                {
                    i++;
                }
            }          
            return -1;
        }

        public static bool CheckAccessKey(string accessKey, out string error)
        {
            int i = 0;
            while (true)
            {
                try
                {
                    var client = OzHfHTTPClient.Get();
                    var request = new HttpRequestMessage(HttpMethod.Get, "https://huggingface.co/api/models");
                    request.Headers.Authorization = new AuthenticationHeaderValue("Bearer", accessKey);
                    var resp = client.Send(request);
                    error = "";
                    if (resp.StatusCode == HttpStatusCode.OK)
                    {
                        return true;
                    }
                    if (resp.StatusCode == HttpStatusCode.Unauthorized)
                    {
                        error = "Invalid access key";
                        return false;
                    }
                }
                catch (Exception ex)
                {
                    i++;
                    if (i > 5)
                    {
                        error = "Error happened while checking access key: " + ex.Message;
                        return false;
                    }
                }
            }
        }

        public static bool GetHuggingFaceRepo(string url, out OzHfHuggingFaceRepository repository, out string error, string? accesstoken = null)
        {
            repository = new OzHfHuggingFaceRepository("", []);
            string? response = null;
            int i = 0;
            while (response == null)
            {
                try
                {
                    var client = OzHfHTTPClient.Get();
                    var request = new HttpRequestMessage(HttpMethod.Get, url);
                    var resp = client.Send(request);
                    if (resp.StatusCode != HttpStatusCode.OK)
                    {
                        error = "Specified repository doesn't exist.";
                        return false;
                    }
                    response = resp.Content.ReadAsStringAsync().Result;
                }
                catch (Exception ex)
                {
                    i++;
                    if (i > 5)
                    {
                        error = "An error happened while reading the HuggingFace repository: " + ex.Message;
                        return false;
                    }
                }
            }

            i = 0;
            while (true)
            {
                try
                {
                    var hFRepoDTO = JsonSerializer.Deserialize<OzHfRepoDTO>(response);
                    if (hFRepoDTO == null)
                    {
                        throw new JsonException("Generated JSON is null.");
                    }
                    var id = hFRepoDTO.Id;
                    var files = hFRepoDTO.Siblings.Select(r => r.Rfilename);
                    repository = new OzHfHuggingFaceRepository(id, files);
                    var client = OzHfHTTPClient.Get();
                    var request = new HttpRequestMessage(HttpMethod.Get, "https://huggingface.co/" + id + "/resolve/main/.gitattributes");

                    if (accesstoken != null)
                    {
                        request.Headers.Authorization = new AuthenticationHeaderValue("Bearer", accesstoken);
                    }
                    var resp = client.Send(request);
                    error = "";
                    if (resp.StatusCode == HttpStatusCode.Unauthorized)
                    {
                        error = "The entered repository is gated. Please provide an HuggingFace access token.";
                        return false;
                    }
                    if (resp.StatusCode == HttpStatusCode.OK)
                    {
                        return true;
                    }
                    if (resp.StatusCode == HttpStatusCode.Forbidden)
                    {
                        error = "The HuggingFace user associated with the provided access key doesn't have access to this gated model.";
                        return false;
                    }
                }
                catch (Exception ex)
                {
                    i++;
                    if (i > 5)
                    {
                        error = "Error happened while gathering information about the HuggingFace repository: " + ex.Message + ":" + response;
                        return false;
                    }
                }
            }
        }
    }
}
